/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.activemq.artemis.tests;

import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.junit.Assert;

import java.net.HttpURLConnection;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.Authenticator;
import java.net.PasswordAuthentication;
import java.net.URL;
import java.util.Base64;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;

// import org.apache.activemq.artemis.api.core.TransportConfiguration;
import org.apache.activemq.artemis.api.jms.ActiveMQJMSClient;
import org.json.JSONObject;


/**
 * Testcase for ENTMQBR-782.
 * Very basic test that runs five producers and five consumers, all connected to
 * the same queue on *external* broker.
 * With multiple producers and consumers the JMX queue counters for
 * "Messages added" and "Messages acknowledged" get out of sync after a while.
 * 
 * At the end of the test it queries the queue's JMX stats for 
 * "Messages added" and "Messages acknowledged". They should have an equal count.
 * If not, the test fails.
 * 
 * Please update BROKERURL and JOLOKIAURL prior to running the test
 */
public class IncorrectQueueStatsTest {

    protected final static String DESTNAME = "TEST.QUEUE";
    protected final static String BROKERURL = "tcp://192.168.99.100:32814";
    protected final static String JOLOKIAURL = "http://192.168.99.100:32817/hawtio/jolokia/read/";
    protected final static String BROKERNAME = "testBroker";
    protected final static int NUM_MSGS = 5000;
    protected final static int NUM_PRODUCERS = 5;
    protected final static int NUM_CONSUMERS = 5;
    protected final static int CONSUMER_SLEEP = 5;
    protected final static int PRODUCER_SLEEP = 5;
    public static CountDownLatch producersFinished = new CountDownLatch(NUM_PRODUCERS);
    public static CountDownLatch consumersFinished = new CountDownLatch(NUM_CONSUMERS );
    protected static Logger LOG = LoggerFactory.getLogger(IncorrectQueueStatsTest.class);


    @Test
    public void testTwoQueueProducersConsumersJms() throws Exception {
       LOG.info("Starting testTwoQueueProducersConsumersJms().");
       ExecutorService producers = Executors.newFixedThreadPool(NUM_PRODUCERS);
       ExecutorService consumers = Executors.newFixedThreadPool(NUM_CONSUMERS);
       createDefaultProducers(producers);
       Thread.sleep(1000);
       createDefaultConsumers(consumers, false);

       producersFinished.await(180, TimeUnit.SECONDS);
       consumersFinished.await(180, TimeUnit.SECONDS);
       LOG.info("All producers have terminated.");
       LOG.info("All consumers have terminated.");

       String jmxContextQueue = 
           "org.apache.activemq.artemis:broker=\"" + 
           BROKERNAME + 
           "\",component=addresses,address=\"" + 
           DESTNAME + 
           "\",subcomponent=queues,routing-type=\"anycast\",queue=\"" +
           DESTNAME + "\"";

       int messagesAcked = getJMXProperty(jmxContextQueue + "/MessagesAcknowledged");
       LOG.info("Messages acked: " + messagesAcked);

       int messagesAdded = getJMXProperty(jmxContextQueue + "/MessagesAdded");
       LOG.info("Messages added: " + messagesAdded);

       int messageCount = getJMXProperty(jmxContextQueue + "/MessageCount");
       LOG.info("Messages count: " + messageCount);

       Assert.assertEquals(0, messageCount);
       Assert.assertEquals(messagesAdded, messagesAcked);
       LOG.info("Done testTwoQueueProducersConsumersJms().");
    }


    protected void createDefaultConsumers(ExecutorService consumers, boolean isTopic) {
        for (int i = 0; i < NUM_CONSUMERS; i++) {
            LOG.info("Creating consumer for destination " + DESTNAME);
            Consumer consumer = new Consumer(DESTNAME, isTopic);
            Thread t = new Thread(consumer);
            t.setName(DESTNAME);
            consumers.submit(t);

            // wait for consumer to signal it has fully initialized
            synchronized(consumer.init) {
                try {
                    consumer.init.wait();
                } catch (InterruptedException e) {
                    LOG.error(e.toString(), e);
                }
            }
        }
    }


    protected void createDefaultProducers(ExecutorService producers) {
        for (int i = 0; i < NUM_PRODUCERS; i++) {
            LOG.info("Creating producer for destination " + DESTNAME );
            Thread producer = new Thread(new Producer(DESTNAME, false));
            producer.setName("Producer for " + DESTNAME + i);
            producers.submit(producer);
        }
    }


    /** 
     * Requests JMX property from Artemis server based on given url context 
     * path. Data is retrieved via jolokia.
     * Uses low level java.net.HttpUrlConnection to make HTTP request as
     * commons-httpclient does not accept " in url and the jolokia url contains
     * a few of this character.
     * @param jmxContext - the relative url path (jmxContext)
     * @return value of the requested JMX property as int
     * @throws Exception
     */
    protected int getJMXProperty(String jmxContext) throws Exception {
        if (jmxContext == null || jmxContext.length() == 0)
            return -1;

        URL url = new URL(JOLOKIAURL + jmxContext);
        HttpURLConnection httpclient = (HttpURLConnection)url.openConnection();
        httpclient.setRequestMethod("GET");
        httpclient.setRequestProperty("User-Agent", "Mozilla/5.0");

        // does not seem to work for setting basic auth credentials
        Authenticator.setDefault (new Authenticator() {
            @Override
            protected PasswordAuthentication getPasswordAuthentication() {
                return new PasswordAuthentication ("admin", "admin".toCharArray());
            }
        });

        // set HTTP basic auth credentials
        String userpass = "admin:admin";
        String basicAuth = "Basic " + new String(Base64.getEncoder().encode(userpass.getBytes()));
        httpclient.setRequestProperty ("Authorization", basicAuth);

        int responseCode = httpclient.getResponseCode();
        StringBuffer response = null;
        if (responseCode == HttpURLConnection.HTTP_OK) {
            BufferedReader in = new BufferedReader(new InputStreamReader(httpclient.getInputStream()));
            String inputLine;
            response = new StringBuffer();

            while ((inputLine = in.readLine()) != null) {
                response.append(inputLine);
            }
            in.close();

            LOG.debug(response.toString());
        } else {
            LOG.error("GET request failed with response code " + httpclient.getResponseMessage());
        }
        // parse request as JSON Object
        JSONObject json = new JSONObject(response.toString());
        Integer num = (Integer) json.get("value");
        return num.intValue();
    }
}


/**
 * Simple consumer class that is supposed to run in an extra thread and 
 * uses JMS API to consume a number of messages.
 * Some consumer properties are configured upper class
 */
@SuppressWarnings("MultipleTopLevelClasses")
class Consumer implements Runnable {

    protected String queueName = "";
    boolean isTopic = false;
    public final Object init = new Object();
    protected static Logger LOG = LoggerFactory.getLogger(Consumer.class);

    public Consumer(String destName, boolean topic) {
        this.isTopic = topic;
        this.queueName = destName;
    }


    @Override
    public void run() {
        Connection connection = null;
        Session session = null;
        MessageConsumer consumer = null;
        Destination dest = null;
        try {
            ConnectionFactory cf = ActiveMQJMSClient.createConnectionFactory(IncorrectQueueStatsTest.BROKERURL, "standard");

            //Next we create a JMS connection using the connection factory:
            connection = cf.createConnection();

            //And we create a non transacted JMS Session, with AUTO\_ACKNOWLEDGE
            //acknowledge mode:
            session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);

            if (isTopic) {
                dest = session.createTopic(IncorrectQueueStatsTest.DESTNAME);
            } else {
                dest = session.createQueue(IncorrectQueueStatsTest.DESTNAME);
            }

            //And we create a MessageConsumer which will consume orders from the
            //queue:ao
            if (isTopic) {
                consumer = session.createDurableSubscriber((Topic) dest, "subscriber-id-1");
            } else {
                consumer = session.createConsumer(dest);
            }
            //We make sure we start the connection, or delivery won't occur on it:
            connection.start();

            synchronized(init) {
                init.notifyAll();
            }

            //And we consume the message:
            for (int i=0; i < IncorrectQueueStatsTest.NUM_MSGS; i++) {
                TextMessage receivedMessage = (TextMessage)consumer.receive();
                LOG.debug("Got msg: " + receivedMessage.getText());
                Thread.sleep(IncorrectQueueStatsTest.CONSUMER_SLEEP);

            }
        } catch(Exception e) {
            LOG.error("Error in Consumer: " + e.getMessage());
            return;
        }
        finally {
            try {
                IncorrectQueueStatsTest.consumersFinished.countDown();
                if (consumer != null)
                    consumer.close();
                if (session != null)
                    session.close();
                if (connection != null)
                    connection.close();
            } catch (Exception ex) {
                LOG.error("Error closing down JMS objects: " + ex.getMessage());
            }
        }
    }
}

/**
 * Simple producer class that is supposed to run in an extra thread and uses 
 * the JMS API to produce a number of messages.
 * Some consumer properties are configured upper class
 */
@SuppressWarnings("MultipleTopLevelClasses")
class Producer implements Runnable{

    protected static Logger LOG = LoggerFactory.getLogger(Producer.class);
    protected int timeToLive=60000;
    protected String destName = "";
    protected boolean isTopic = false;

    public Producer(String dest, boolean isTopic) {
        this.destName = dest;
        this.isTopic = isTopic;
    }


    @Override
    public void run() {
        Connection connection = null;
        Session session = null;
        MessageProducer producer = null;
        Destination dest = null;
        try {
            ConnectionFactory cf = ActiveMQJMSClient.createConnectionFactory(IncorrectQueueStatsTest.BROKERURL, "standard");

            //Next we create a JMS connection using the connection factory:
            connection = cf.createConnection();

            //And we create a non transacted JMS Session, with AUTO\_ACKNOWLEDGE
            //acknowledge mode:
            session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);

            if (isTopic) {
                dest = session.createTopic(IncorrectQueueStatsTest.DESTNAME);
            } else {
                dest = session.createQueue(IncorrectQueueStatsTest.DESTNAME);
            }

            //We make sure we start the connection, or delivery won't occur on it:
            connection.start();

            producer = session.createProducer(dest);
            //We create a simple TextMessage and send it:
            TextMessage message = session.createTextMessage("This is a message.");

            for (int i=0; i < IncorrectQueueStatsTest.NUM_MSGS; i++) {
                producer.send(message);
                LOG.debug("Message sent.");
                Thread.sleep(IncorrectQueueStatsTest.PRODUCER_SLEEP);
            }
        } catch(Exception e) {
            LOG.error("Error in Consumer: " + e.getMessage());
            return;
        }
        finally {
            try {
                IncorrectQueueStatsTest.producersFinished.countDown();
                if (producer != null)
                   producer.close();
                if (session != null)
                    session.close();
                if (connection != null)
                    connection.close();
            } catch (Exception ex) {
                LOG.error("Error closing down JMS objects: " + ex.getMessage());
            }
        }
    }
}

